from opentelemetry.instrumentation.requests import RequestsInstrumentor

from ..utils import Utils

from cisco_opentelemetry_specifications import SemanticAttributes


def get_active_span_for_call_wrapper():
    def get_active_span_for_call(span, response) -> None:
        if not span.is_recording():
            return

        if hasattr(response, "request"):
            Utils.add_flattened_dict(
                span,
                SemanticAttributes.HTTP_REQUEST_HEADER,
                getattr(response.request, "headers", dict()),
            )

            Utils.set_payload(
                span,
                SemanticAttributes.HTTP_REQUEST_BODY,
                getattr(response.request, "body", str()),
            )

        Utils.add_flattened_dict(
            span,
            SemanticAttributes.HTTP_RESPONSE_HEADER,
            getattr(response, "headers", dict()),
        )

        Utils.set_payload(
            span,
            SemanticAttributes.HTTP_RESPONSE_BODY,
            getattr(response, "content", bytes()),
        )

    return get_active_span_for_call


def name_callback(method, url) -> str:
    return method + " " + url


class RequestsInstrumentorWrapper(RequestsInstrumentor):
    def __init__(self):
        super().__init__()

    def _instrument(self, **kwargs) -> None:
        super()._instrument(
            tracer_provider=kwargs.get("tracer_provider"),
            span_callback=get_active_span_for_call_wrapper(),
            name_callback=name_callback,
        )

    def _uninstrument(self, **kwargs) -> None:
        super()._uninstrument()



import json
import grpc
import logging
from typing import MutableMapping
from collections import OrderedDict

from wrapt import wrap_function_wrapper as _wrap
from typing import Any
from google.protobuf.json_format import MessageToDict
from opentelemetry import trace
from opentelemetry.instrumentation.grpc import (
    GrpcInstrumentorServer,
    GrpcInstrumentorClient,
    _server,
    _client,
)
from opentelemetry.propagate import inject
from opentelemetry.propagators.textmap import Setter
from opentelemetry.instrumentation.grpc.version import __version__
from opentelemetry.instrumentation.grpc.grpcext import intercept_channel
from ..utils import Utils
from opentelemetry.instrumentation.grpc._utilities import RpcInfo
from opentelemetry.trace.status import Status, StatusCode
from cisco_telescope.instrumentations import utils

from cisco_opentelemetry_specifications import SemanticAttributes
from opentelemetry.semconv.trace import SpanAttributes


# code was taken from github commit sha: f7eb9673bca5d6fb4d16040e8ac28053225ad302
# https://github.com/hypertrace/pythonagent/pull/262


class GrpcInstrumentorServerWrapper(GrpcInstrumentorServer):
    """wrapper around OTel grpc:server instrumentor class"""

    def __init__(self):
        super().__init__()
        self._original_wrapper_func = None

    def instrument(self, **kwargs) -> None:
        """instrument grpc:server"""
        super().instrument()

    def uninstrument(self, **kwargs) -> None:
        """disable grpc:server instrumentation."""
        super().uninstrument()

    # Internal enable wrapper instrumentation
    def _instrument(self, **kwargs) -> None:
        """Enable wrapper instrumentation internal call"""
        super()._instrument(**kwargs)
        self._original_wrapper_func = grpc.server

        def server_wrapper(*args, **kwargs) -> None:
            kwargs["interceptors"] = [server_interceptor_wrapper(self)]
            return self._original_wrapper_func(*args, **kwargs)

        grpc.server = server_wrapper

    # Internal disable wrapper instrumentation
    def _uninstrument(self, **kwargs) -> None:
        """Disable wrapper instrumentation internal call"""
        super()._uninstrument(**kwargs)


# The main entry point for a wrapper around the OTel grpc:client instrumentation module
class GrpcInstrumentorClientWrapper(GrpcInstrumentorClient):
    def __init__(self):
        super().__init__()

    def _instrument(self, **kwargs) -> None:
        """Internal initialize instrumentation"""
        logging.debug("Instrumenting")
        for ctype in self._which_channel(kwargs):
            _wrap(
                "grpc",
                ctype,
                self.wrapper_fn_wrapper,
            )

    def _uninstrument(self, **kwargs) -> None:
        """Internal disable instrumentation"""
        super()._uninstrument(**kwargs)

    def wrapper_fn_wrapper(self, original_func, instance, args, kwargs) -> None:
        """Wrap function for initializing the request handler"""
        channel = original_func(*args, **kwargs)
        tracer_provider = kwargs.get("tracer_provider")
        return intercept_channel(
            channel,
            client_interceptor_wrapper(self, tracer_provider=tracer_provider),
        )


def server_interceptor_wrapper(gisw, tracer_provider=None):
    """Helper function to set interceptor."""
    tracer = trace.get_tracer(__name__, __version__, tracer_provider)
    return OpenTelemetryServerInterceptorWrapper(tracer, gisw)


def client_interceptor_wrapper(gicw, tracer_provider):
    """Helper function to set interceptor."""
    tracer = trace.get_tracer(__name__, __version__, tracer_provider)
    return OpenTelemetryClientInterceptorWrapper(tracer, gicw)


class _OpenTelemetryWrapperServicerContext(_server._OpenTelemetryServicerContext):
    """
    Wrapper around Server-side telemetry context
    grpc:server telemetry context
    """

    def __init__(self, servicer_context, active_span):
        super().__init__(servicer_context, active_span)
        self._response_headers = ()

    def set_trailing_metadata(self, *args, **kwargs) -> None:
        """Override trailing metadata(response headers) method.
        Allows us to capture the response headers"""
        self._response_headers = args
        return self._servicer_context.set_trailing_metadata(*args, **kwargs)

    def get_trailing_metadata(self) -> tuple:
        """Return response headers"""
        return self._response_headers


class OpenTelemetryServerInterceptorWrapper(_server.OpenTelemetryServerInterceptor):
    """Wrapper around server-side interceptor"""

    def __init__(self, tracer, gisw):
        super().__init__(tracer)
        self._gisw = gisw

    def intercept_service(self, continuation, handler_call_details):
        logging.debug("Intercepting")

        def telemetry_wrapper(behavior, request_streaming, response_streaming):
            def telemetry_interceptor(request_or_iterator, context) -> Any:
                # handle streaming responses specially
                if response_streaming:
                    return self._intercept_server_stream(
                        behavior,
                        handler_call_details,
                        request_or_iterator,
                        context,
                    )

                span = context._active_span

                Utils.set_payload(
                    span,
                    SemanticAttributes.RPC_REQUEST_BODY,
                    json.dumps(MessageToDict(request_or_iterator)),
                )

                Utils.add_flattened_dict(
                    span,
                    SemanticAttributes.RPC_REQUEST_METADATA,
                    dict(handler_call_details.invocation_metadata),
                )

                try:
                    context = _OpenTelemetryWrapperServicerContext(context, span)
                    # Capture response
                    response = behavior(request_or_iterator, context)

                except Exception as error:
                    # Bare exceptions are likely to be gRPC aborts, which
                    # we handle in our context wrapper.
                    # Here, we're interested in uncaught exceptions.
                    # pylint:disable=unidiomatic-typecheck
                    logging.exception("Exception in user context call")
                    if type(error) != Exception:
                        span.record_exception(error)
                        raise error

                else:
                    trailing_metadata = context.get_trailing_metadata()
                    if len(trailing_metadata) > 0:
                        trailing_metadata = dict(trailing_metadata[0])
                    else:
                        trailing_metadata = {}

                    response_dict = MessageToDict(response)

                    Utils.set_payload(
                        span,
                        SemanticAttributes.RPC_RESPONSE_BODY,
                        json.dumps(response_dict),
                    )

                    Utils.add_flattened_dict(
                        span,
                        SemanticAttributes.RPC_RESPONSE_METADATA,
                        trailing_metadata,
                    )

                    return response

            return telemetry_interceptor

        return _server._wrap_rpc_behavior(
            continuation(handler_call_details), telemetry_wrapper
        )  # pylint: disable=W0212

    # def _intercept_server_stream(
    #         self, behavior, handler_call_details, request_or_iterator, context
    # ):
    #     print("Entering OpenTelemetryServerInterceptorWrapper.intercept_server_stream().")
    #     with self._set_remote_context(context):
    #         with self._start_span(
    #                 handler_call_details, context, set_status_on_exception=False
    #         ) as span:
    #             context = _OpenTelemetryWrapperServicerContext(context, span)
    #
    #             try:
    #                 yield from behavior(request_or_iterator, context)
    #
    #             except Exception as error:
    #                 # pylint:disable=unidiomatic-typecheck
    #                 if type(error) != Exception:
    #                     span.record_exception(error)
    #                 raise error


class _CarrierSetter(Setter):
    """
    We use a custom setter in order to be able to lower case
    keys as is required by grpc.
    """

    def set(self, carrier: MutableMapping[str, str], key: str, value: str):
        carrier[key.lower()] = value


class OpenTelemetryClientInterceptorWrapper(_client.OpenTelemetryClientInterceptor):
    """Wrapper around client-side interceptor"""

    def __init__(self, tracer, gicw):
        super().__init__(tracer)
        self._gicw = gicw

    def _intercept(self, request, metadata, client_info, invoker):
        # if context.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
        #     return invoker(request, metadata)
        if not metadata:
            mutable_metadata = OrderedDict()
        else:
            mutable_metadata = OrderedDict(metadata)
        with self._start_span(
            client_info.full_method,
            end_on_exit=False,
            record_exception=False,
            set_status_on_exception=False,
        ) as span:
            result = None
            try:
                inject(mutable_metadata, setter=_CarrierSetter())
                metadata = tuple(mutable_metadata.items())

                rpc_info = RpcInfo(
                    full_method=client_info.full_method,
                    metadata=metadata,
                    timeout=client_info.timeout,
                    request=request,
                )

                result = invoker(request, metadata)

                # Add request headers
                Utils.add_flattened_dict(
                    span,
                    SemanticAttributes.RPC_REQUEST_METADATA,
                    utils.Utils.lowercase_items(dict(metadata)),
                )

                # Add request body
                Utils.set_payload(
                    span,
                    SemanticAttributes.RPC_REQUEST_BODY,
                    str(request),  # body
                )

            except Exception as exc:
                if isinstance(exc, grpc.RpcError):
                    span.set_attribute(
                        SpanAttributes.RPC_GRPC_STATUS_CODE,
                        exc.code().value[0],
                    )
                span.set_status(
                    Status(
                        status_code=StatusCode.ERROR,
                        description=f"{type(exc).__name__}: {exc}",
                    )
                )
                span.record_exception(exc)
                raise exc
            finally:
                if not result:
                    span.end()

        return self._trace_result(span, rpc_info, result)

    # def _intercept_server_stream(
    #         self, request_or_iterator, metadata, client_info, invoker
    # ):
    #     print("Entering OpenTelemetryClientInterceptorWrapper.intercept_server_stream().")
    #     if not metadata:
    #         mutable_metadata = OrderedDict()
    #     else:
    #         mutable_metadata = OrderedDict(metadata)
    #
    #     with self._start_span(client_info.full_method) as span:
    #         inject(mutable_metadata, setter=_CarrierSetter())
    #         metadata = tuple(mutable_metadata.items())
    #         rpc_info = RpcInfo(
    #             full_method=client_info.full_method,
    #             metadata=metadata,
    #             timeout=client_info.timeout,
    #         )
    #
    #         if client_info.is_client_stream:
    #             rpc_info.request = request_or_iterator
    #
    #         try:
    #             yield from invoker(request_or_iterator, metadata)
    #         except grpc.RpcError as err:
    #             span.set_status(Status(StatusCode.ERROR))
    #             span.set_attribute(
    #                 SpanAttributes.RPC_GRPC_STATUS_CODE, err.code().value[0]
    #             )
    #             raise err

    # def intercept_stream(
    #     self, request_or_iterator, metadata, client_info, invoker
    # ):
    #     print("Entering OpenTelemetryClientInterceptorWrapper.intercept_stream().")
    #     if client_info.is_server_stream:
    #         return self._intercept_server_stream(
    #             request_or_iterator, metadata, client_info, invoker
    #         )
    #
    #     return self._intercept(
    #         request_or_iterator, metadata, client_info, invoker
    #     )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cisco_telescope',
 'cisco_telescope.cmd',
 'cisco_telescope.instrumentations',
 'cisco_telescope.instrumentations.aiohttp',
 'cisco_telescope.instrumentations.grpc',
 'cisco_telescope.instrumentations.requests']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<4.0.0',
 'argparse>=1.4.0,<2.0.0',
 'cisco-opentelemetry-specifications>=0.0.15,<0.0.16',
 'importlib-metadata>=4.11.3,<5.0.0',
 'opentelemetry-exporter-otlp>=1.10.0,<2.0.0',
 'opentelemetry-instrumentation-aiohttp-client>=0.29b0,<0.30',
 'opentelemetry-instrumentation-grpc==0.29b0',
 'opentelemetry-instrumentation-requests>=0.29b0,<0.30',
 'opentelemetry-instrumentation>=0.29b0,<0.30',
 'opentelemetry-sdk>=1.10.0,<2.0.0',
 'opentelemetry-semantic-conventions>=0.29b0,<0.30',
 'requests>=2.27.1,<3.0.0']

entry_points = \
{'console_scripts': ['telescope = cisco_telescope.cmd.trace:run',
                     'telescope-bootstrap = cisco_telescope.cmd.bootstrap:run']}

setup_kwargs = {
    'name': 'cisco-telescope',
    'version': '0.3.5',
    'description': 'Cisco Distribution for OpenTelemetry',
    'long_description': '# otel-py\n[![PIP Published Version][pip-image]][pip-url]\n[![Apache License][license-image]][license-image]\n[![Coverage][coverage-image]][coverage-url]\n\n![Trace](trace.png)\n\nThis package provides OpenTelemetry-compliant tracing to Python\napplications for the collection of distributed tracing and performance metrics in [Cisco Telescope](https://console.telescope.app/?utm_source=github).\n\n## Contents\n\n- [Installation](#installation)\n  - [Install Packages](#install-packages)\n  - [Install Dependencies](#install-dependencies)\n  - [Library initialization](#library-initialization)\n  - [OpenTelemetry Collector Configuration](#opentelemetry-collector-configuration)\n  - [Existing OpenTelemetry Instrumentation](#existing-opentelemetry-instrumentation)\n- [Supported Runtimes](#supported-runtimes)\n- [Frameworks](#frameworks)\n- [Supported Libraries](#supported-libraries)\n- [Configuration](#configuration)\n- [Getting Help](#getting-help)\n- [Opening Issues](#opening-issues)\n- [License](#license)\n\n## Installation\n\n### Install packages\n> To install Cisco OpenTelemetry Distribution simply run:\n\n```sh\npip install cisco-telescope\n```\n\n### Install dependencies\n> To install all supported instrumentation frameworks run:\n```sh\ntelescope-bootstrap\n```\n\n\n### Library initialization\n> Cisco OpenTelemetry Distribution is activated and instruments the supported libraries once the `tracing.init()` has returned.\n\n```python\nfrom cisco_telescope import tracing\n\ntracing.init(\n  service_name="<your-service-name>",\n  cisco_token="<your-telescope-token>"\n)\n```\n\n### OpenTelemetry Collector Configuration\n\n> By default, Cisco OpenTelemetry Distribution exports data directly to [Cisco Telescope\'s](https://console.telescope.app/?utm_source=github) infrastructure backend.\n> **Existing** OpenTelemetery Collector is supported, the following configuration can be applied\n\n#### Configure custom trace exporter\n\n> Cisco OpenTelemetry Distribution supports configure multiple custom exporters.\n> Example for create OtlpGrpc Span exporter to local OpenTelemetry collector:\n\n```python\nfrom cisco_telescope import tracing, options\n\ntracing.init(\n  service_name="<your-service-name>",\n  cisco_token="<your-telescope-token>",\n  exporters=[\n    options.ExporterOptions(\n      exporter_type="otlp-grpc",\n      collector_endpoint="grpc://localhost:4317"\n    ),\n  ]\n)\n\n```\n\n#### Configure custom OpenTelemetry collector to export trace data to [Cisco Telescope\'s](https://console.telescope.app/?utm_source=github) external collector.\n\n```yaml\ncollector.yaml ...\n\nexporters:\n  otlphttp:\n    traces_endpoint: https://production.cisco-udp.com/trace-collector:80\n    headers:\n      authorization: <Your Telescope Token>\n    compression: gzip\n\n\nservice:\n  pipelines:\n    traces:\n      exporters: [otlphttp]\n```\n\n### Existing OpenTelemetry Instrumentation\n\n> Notice: Only relevant if interested in streaming existing OpenTelemetry workloads.\n> [Cisco Telescope](https://console.telescope.app/?utm_source=github). supports native OpenTelemetery traces.\n```python\nfrom opentelemetry import trace\nfrom opentelemetry.sdk.resources import Resource\nfrom opentelemetry.sdk.trace import TracerProvider\nfrom opentelemetry.sdk.trace.export import BatchSpanProcessor\n\nfrom opentelemetry.exporter.otlp.proto.http.trace_exporter import (\n    OTLPSpanExporter as OTLPHTTPExporter,\n)\n\nprovider = TracerProvider(resource=Resource.create())\ntrace.set_tracer_provider(provider)\n\nhttp_exporter = OTLPHTTPExporter(\n  endpoint="https://production.cisco-udp.com/trace-collector:80",\n  headers= {\n    "authorization": "<Your Telescope Token>",\n  },\n)\n\nprocessor = BatchSpanProcessor(http_exporter)\nprovider.add_span_processor(processor)\n```\n\n## Supported Runtimes\nCisco OpenTelemetry Distribution supports Python 3.6+\n\n## Frameworks\n\n> Cisco OpenTelemetry Python Distribution is extending Native OpenTelemetry, supported frameworks [available here](https://github.com/open-telemetry/opentelemetry-python-contrib/tree/main/instrumentation).\n\n## Supported Libraries\n\n> Cisco OpenTelemetry Python Distribution is extending Native OpenTelemetry, supported libraries [available here](https://github.com/open-telemetry/opentelemetry-js-contrib/tree/main/metapackages/auto-instrumentations-node#supported-instrumentations).\n\nCisco OpenTelemetry Python Distribution provides out-of-the-box instrumentation (tracing) and advanced **payload collections** for many popular frameworks and libraries.\n\n| Library  | Extended Support Version    |\n|----------|-----------------------------|\n| requests | Fully supported             |\n| aiohttp  | Fully supported             |\n| grpc     | Client/Server Unary support |\n\n## Configuration\n\nAdvanced options can be configured as a parameter to the init() method:\n\n| Parameter        | Env                    | Type    | Default       | Description                                                       |\n|------------------|------------------------| ------- |---------------| ----------------------------------------------------------------- |\n| cisco_token      | CISCO_TOKEN            | string  | -             | Cisco account token                                               |\n| service_name     | OTEL_SERVICE_NAME      | string  | `application` | Application name that will be set for traces                      |\n| debug            | CISCO_DEBUG            | string  | `False`       | Debug logs                                                        |\n| payloads_enabled | CISCO_PAYLOADS_ENABLED | string  | `True`        | Debug logs                                                        |\n\nExporter options\n\n| Parameter          | Env                     | Type                | Default                                               | Description                                                                                                                                 |\n|--------------------| ----------------------- | ------------------- | ----------------------------------------------------- | ------------------------------------------------------------------------------------------------------------------------------------------- |\n| collector_endpoint | OTEL_COLLECTOR_ENDPOINT | string              | `https://production.cisco-udp.com/trace-collector:80` | The address of the trace collector to send traces to                                                                                        |\n| type               | OTEL_EXPORTER_TYPE      | string              | `otlp-http`                                           | The exporter type to use (Currently only `otlp-http` are supported). Multiple exporter option available via init function see example below |\n\n## Getting Help\n\nIf you have any issue around using the library or the product, please don\'t hesitate to:\n\n- Use the [documentation](https://docs.telescope.app).\n- Use the help widget inside the product.\n- Open an issue in GitHub.\n\n## License\n\nProvided under the Apache 2.0. See LICENSE for details.\n\nCopyright 2022, Cisco\n\n[pip-url]: https://pypi.org/project/cisco-telescope/\n[pip-image]: https://img.shields.io/github/v/release/cisco-open/otel-py?include_prereleases&style=for-the-badge\n[license-url]: https://github.com/https://github.com/cisco-open/otel-py/blob/main/LICENSE\n[license-image]: https://img.shields.io/badge/license-Apache_2.0-green.svg?style=for-the-badge\n[coverage-url]: https://codecov.io/gh/cisco-open/otel-py/branch/main/\n[coverage-image]: https://img.shields.io/codecov/c/github/cisco-open/otel-py?style=for-the-badge\n',
    'author': 'Cisco Epsagon Team',
    'author_email': 'support@epsagon.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cisco-open/otel-py',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Copyright The Cisco Authors

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import options
from . import consts

from opentelemetry.sdk.trace.export import ConsoleSpanExporter
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter as OTLPGrpcExporter,
)
from opentelemetry.exporter.otlp.proto.http.trace_exporter import (
    OTLPSpanExporter as OTLPHTTPExporter,
)


def _set_exporter(exporter: options.ExporterOptions, opt: options.Options):
    if exporter.exporter_type == consts.GRPC_EXPORTER_TYPE:
        return OTLPGrpcExporter(
            endpoint=exporter.collector_endpoint, headers=exporter.custom_headers
        )

    elif exporter.exporter_type == consts.HTTP_EXPORTER_TYPE:
        return OTLPHTTPExporter(
            endpoint=exporter.collector_endpoint, headers=exporter.custom_headers
        )
    elif exporter.exporter_type == consts.CONSOLE_EXPORTER_TYPE:
        return ConsoleSpanExporter(service_name=opt.service_name)
    else:
        raise ValueError("Unsupported exported type")


def init_exporters(opt: options.Options):
    return [_set_exporter(exporter, opt) for exporter in opt.exporters]
